#!/usr/bin/suidperl -T
#
# File: qmail-scanner-queue.pl
# Version: 1.13
#
# Author: Jason L. Haar <jhaar@users.sourceforge.net>
# 
# This file was auto-generated by:
#
# ./configure --spooldir /var/spool/qmailscan --qmaildir /var/qmail --bindir /var/qmail/bin --qmail-queue-binary /var/qmail/bin/qmail-queue  --admin root --domain some.domain.com --notify sender,admin --local-domains some.domain.com --silent-viruses klez --lang en_GB --debug 1 --unzip 1 --add-dscr-hdrs 0 --archive 0 --redundant 0 --log-details  --fix-mime 1  --scanners "auto"
# 
# Description: This is a replacement/add-on for Qmail 1.0.3's qmail-queue.
# It can call several blocking programs - such as virus scanners - on every 
# SMTP-received Email message, checking for viruses and blocked filenames, 
# only allowing the message to continue if it passes the tests.
#
#   Copyright (C) 1999,2000,2001 the people mentioned above
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 1, or (at your option)
#   any later version. See <URL:http://www.gnu.org/copyleft/gpl.html>
#   for a copy.
# 
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   The software is provided as is. Please bear in mind that we have 
#   done this in my spare time. While it is as accurate as we could 
#   make it there is a reasonable chance that there are mistakes 
#   somewhere in here. If you email me and tell me about them, I will 
#   be happy to fix them but I can't take responsibility for your system. 
#   Basically use this at your own risk. 
#
#####################################################################

#####################################################################
##
## Required Packages
##
## Qmail-1.03
## Perl 5.005_03+
## Maildrop-0.73
## Bruce Guenter's QMAILQUEUE patch <URL:http://www.qmail.org/qmailqueue-patch>
## Perl module Time::HiRes and DB_File
##
##
## So-far tested Virus scanners:
##     Trend's Virus scanner for Linux
##     MacAfee's (NAI's) virus scanner for Linux
##     Sophos's virus scanner for Linux
##     H+BEDV's antivir scanner for Linux
##     F-Secure's fsav scanner for Linux
##     P-Prot for Linux
##     Sophie (daemonized Sophos scanner)
##     Trophie (daemonized Trend scanner)
## 
#####################################################################

#####################################################################
##
## Site-specific config
##
#####################################################################


delete @ENV{qw(IFS CDPATH ENV BASH_ENV QMAILMFTFILE QMAILINJECT)};

use strict 'vars', 'subs';

use Sys::Syslog qw(:DEFAULT setlogsock);
setlogsock('unix');

our $VERSION="1.13";

#Mail header to add to each scanned message to report stuff in...
#Default is to not generate them ($descriptive_hdrs = 0) - as  that
#info is also in the Received: headers...
our $descriptive_hdrs=0;
our $V_HEADER="X-Qmail-Scanner";


#From: line  information used when making reports
our $V_FROM='root@some.domain.com';
our $V_FROMNAME='System Anti-Virus Administrator';

# Address carbon-copied on any virus reports
our $QUARANTINE_CC='root@some.domain.com';

#Array of local domains that are checked against for
#deciding whether or not to send recipient alerts to
our @local_domains_array=('some.domain.com');

# Array of virus that we want to inform the sender of.
our @silent_viruses_array=('klez');


#Array of virus scanners used must point to subroutines
our @scanner_array=();

#Addresses that should be alerted of any quarantined Email
our $NOTIFY_ADDRS='sender,admin';

#Try to fix bad MIME messages before passing to MIME unpacker
our $BAD_MIME='1';

# The full path to qmail programs we'll need.
our $qmailinject = '/var/qmail/bin/qmail-inject';
our $qmailqueue  = '/var/qmail/bin/qmail-queue';

# What directory to use for storing temporary files.
our $scandir = '/var/spool/qmailscan';

#What maildir folder to store working files in
our $wmaildir='working';

#What maildir folder to store failed messages in (for cronjob scan)
our $fmaildir='failed';

#What maildir folder to store quarantine in
our $vmaildir='quarantine';


#What maildir folder to archive received Email in instead of deleting
our $archiveit='0';
our $archivedir='archives';

#Name of file in $scandir where debugging output goes
our $debuglog="qmail-queue.log";

#Name of file where quarantine reports go (for long-term storage)
our $quarantinelog="quarantine.log";

#Generate nice random filename
our $hostname='some.domain.com'; #could get via call I suppose...

#If you trust the virus scanners handling of mbox format itself
#you may want to let it have a go at the "raw" message, and original
#zip files if present
our $redundant_scanning='0';

#If you want to log via file/syslog information of all Email
# that passes through your system (from/to/subj/size/attachments)
our $log_details="0";


#Full path to file in which virus-scanner versioning info is kept
our $versionfile="$scandir/qmail-scanner-queue-version.txt";

#DB file (without extension) where bad filenames are kept.
# You edit $db_filename.txt, and "qmail-scanner-queue.pl -g" generates $db_filename.db
our $db_filename="$scandir/quarantine-attachments";

#What locale is used on this system
#$sys_locale="LOCALE";

#Full paths to binaries used within this script follow - small performance
#improvement :-)


our $mimeunpacker_binary='/usr/bin/reformime ';
our $unzip_binary='/usr/bin/unzip';
our $unzip_options='-Pxxxxx';
our $tnef_binary='';
our $rm_binary='/bin/rm';
our $grep_binary='/bin/grep';
our $find_binary='/usr/bin/find';
our $uudecode_binary='/usr/bin/uudecode';
our $uudecode_pipe='-o -';


our $uvscan_binary='';
our $csav_binary='';
our $sweep_binary='';
our $sophie_binary='';
our $trophie_binary='';
our $iscan_binary='';
our $hbedv_binary='';
our $avp_binary='';
our $avpdaemon_binary='';
our $fprot_binary='';
our $fsecure_binary='';
our $inocucmd_binary='';
our $ravlin_binary='';
our $clamscan_binary='';
our $spamc_binary='';
our $spamc_options='';
our $spamc_subject='';
our $spamassassin_binary='';
our ($sa_comment,$tag_score);

$ENV{'PATH'}='/bin:/usr/bin';

our $SCANINFO='';

our $MAX_FILE_LENGTH=100;
our $MAX_NUM_HDRS=140;
our $QE_LEN=16; 

#Maximum amount of time we allow Q-S to run before returning
# a temp failure. This is so remote SMTP servers don't get confused
# over whether or not they have delivered to a SMTP server
# that's refused to say "OK" for over an hour...
# We'll default to 20 minutes. If the scanner loop takes more than 20 
# minutes to scan the message, then something *must* be wrong with the
# scanner. 
our $MAXTIME=20*60;

#Finally, are you sure your virus scanners can unpack zip files? 
#McAfee's doesn't!
our $force_unzip=1;

#Descriptive string to use in generated Email
our $destring="virus";

#Want debugging? Enable this and read $scandir/qmail-queue.log
our $DEBUG='1';

require "/etc/qmail-scanner.conf";

#####################################################################
## 
## End of site-specific settings
##
#####################################################################

my($qsmsgid);
$qsmsgid=tolower("$V_HEADER-message-id");

my @uufile_list = ();
my @zipfile_list = ();

#Want microsec times for debugging
use Time::HiRes qw( usleep ualarm gettimeofday tv_interval );
use POSIX;

use vars qw/ $opt_v $opt_h $opt_g $opt_r $opt_z/;

use Getopt::Std;

#my ($opt_v,$opt_h,$opt_g,$opt_r,$opt_z);

getopts('vhgrz');

my ($start_time,$last_time);
$start_time = $last_time = [gettimeofday];

(my $prog=$0)=~s/^.*\///g;

if ( $opt_h ) {
  print "

$prog

    -h - This help
    -v - show details about this install. 
         Please include in any bug reports.
    -z - gather virus scanner/DAT versions 
         and cleanup old temp files
    -g - generate perlscanner database
    -r - read from perlscanner database\n";
  exit;
}


if ( $opt_g || $opt_r) {
  &generate_quarantine_db;
    exit 0;
}

if ( $opt_v ) {
  &show_version;
  exit 0;
}

chdir($scandir);
umask(0077);
my $quarantine_event=0;

my $file_id = &uniq_id();

#For security reasons, tighten the follow vars...
$ENV{'SHELL'} = '/bin/sh' if exists $ENV{SHELL};
$ENV{'TMP'} = $ENV{'TMPDIR'} = "$scandir/$file_id";
#$ENV{'QMAILSUSER'} = $ENV{'QMAILSHOST'} = '';



if ($mimeunpacker_binary =~ /reformime/) {
  $mimeunpacker_binary .= " -x$scandir/$file_id/";
}

  
#Get current timestamp for logs
my ($sec,$min,$hour,$mday,$mon,$year);
($sec,$min,$hour,$mday,$mon,$year) = localtime(time);
my $nowtime = sprintf "%02d/%02d/%02d %02d:%02d:%02d", $mday, $mon+1, $year+1900, $hour, $min, $sec;

my ($smtp_sender,$remote_smtp_ip,$uid);

$uid=$<;

if ($DEBUG ) {
  open(LOG,">>$scandir/$debuglog");
  select(LOG);$|=1;
  &debug("+++ starting debugging for process $$ by uid=$uid at $nowtime");
}

&debug("setting UID to EUID so subprocesses can access files generated by this script");
$< = $>;            # set real to effective uid
#$( = $);            # set real to effective gid

&debug("program name is $prog, version $VERSION");
if ($opt_z) {
  &scan_queue;
  exit 0;
}


&scanner_info;

my ($smtp_sender,$remote_smtp_ip);
if ($ENV{'TCPREMOTEIP'}) {
  $smtp_sender="via smtp from $ENV{'TCPREMOTEIP'}";
  $remote_smtp_ip=$ENV{'TCPREMOTEIP'};
  &debug("incoming SMTP connection from $smtp_sender");
} else {
  $smtp_sender="via local process $$";
  $remote_smtp_ip='127.0.0.1';
  &debug("incoming pipe connection from $smtp_sender");
}

my (%headers );
my ($altered_subject, $HEADERS, $env_returnpath, $returnpath);
my ($env_recips, $recips, $one_recip);
my ($alarm_status,$elapsed_time,$msg_size,$file_desc);
my ($description,$quarantine_description,$illegal_mime);


&working_copy;

#Now alarm this area so that hung networks/virus scanners don't cause 
#double-delivery...

eval {
  $SIG{ALRM} = sub { die "Maximum time exceeded. Something cannot handle this message." };
  alarm $MAXTIME;

  &deconstruct_msg if (!$quarantine_event);
  
  
  #Now unset env var QMAILQUEUE so any further Email's sent don't
  #go through the Qmail-Scanner again
  &debug("unsetting QMAILQUEUE env var");
  delete $ENV{'QMAILQUEUE'};
  
  #This SMTP session is incomplete until we see dem envelope headers!
  &grab_envelope_hdrs;
  &debug("from=$headers{'from'},subj=$headers{'subject'}, $qsmsgid=$headers{$qsmsgid} $smtp_sender");

  #Add envelope details to headers array so that they can be matched within
  #perlscanner.
  #Note how they're uppercase cf the message headers which are all forced
  #lowercased. This is to ensure no-one can override them...

  $headers{'MAILFROM'}=$returnpath;
  $headers{'RCPTTO'}=$recips;
  $headers{'TCPREMOTEIP'}=$remote_smtp_ip;

  #Now, start the scanners!
  if (!$quarantine_event) {
    &init_scanners;
  }
  if ($quarantine_event) {
    &debug("unsetting TCPREMOTEIP env var");
    delete $ENV{'TCPREMOTEIP'};
    &email_quarantine_report;
  } else {
    &qmail_parent_check;
    &qmail_requeue($env_returnpath,$env_recips,"$scandir/$wmaildir/new/$file_id"); 
  }
  alarm 0;
};

$alarm_status=$@;
if ($alarm_status and $alarm_status ne "" ) { 
  if ($alarm_status eq "Maximum time exceeded. Something cannot handle this message.") {
    &tempfail("ALARM: taking longer than $MAXTIME secs. Requeuing...");
  } else {
    &tempfail("Requeuing: $alarm_status");
  }
}


#Msg has been delivered now, so don't want hangs in this part
#to affect delivery

if ($log_details) {
  if ($recips =~ /\,/) {
    for $one_recip (split(/\,/,$recips)) {
      &log_msg("qmail-scanner",($quarantine_event ne "0" ? $quarantine_event : "Clear:$tag_score"),$elapsed_time,$msg_size,$returnpath,$one_recip,$headers{'subject'},$headers{$qsmsgid},$file_desc);
    }
  } else {
    #Only one recip
    &log_msg("qmail-scanner",($quarantine_event ne "0" ? $quarantine_event : "Clear:$tag_score"),$elapsed_time,$msg_size,$returnpath,$recips,$headers{'subject'},$headers{$qsmsgid},$file_desc);
  }
}
&cleanup;

($sec,$min,$hour,$mday,$mon,$year) = localtime(time);
$nowtime = sprintf "%02d/%02d/%02d %02d:%02d:%02d", $mday, $mon+1, $year+1900, $hour, $min, $sec;

&debug("all finished. Total of ",tv_interval ($start_time, [gettimeofday])," secs");
exit 0;

############################################################################
# Error handling
############################################################################

#Generate uniq identifiers
sub uniq_id {
  return "$hostname" . time . __LINE__ . $$;
}


# Fail with the given message and a temporary failure code.
sub tempfail {
  syslog('mail|info',"$V_HEADER-$VERSION:[$file_id] @_");
  if ($log_details ne "syslog") {
    warn "$V_HEADER-$VERSION:[$file_id] ",@_, "\n";
  }
  $nowtime = sprintf "%02d/%02d/%02d %02d:%02d:%02d", $mday, $mon+1, $year+1900, $hour, $min, $sec;
  &debug("tempfail: $V_HEADER-$VERSION: ",@_);
  close(LOG);
  &cleanup;
  exit 111;
}

sub debug {
  print LOG "$nowtime:$$: ",@_,"\n" if ($DEBUG);
}

sub working_copy {
  my ($hdr,$last_hdr,$value,$num_of_headers);
  select(STDIN); $|=1;
  
  &debug("w_c: mkdir $scandir/$file_id");
  mkdir("$scandir/$file_id",0700);
  chdir("$scandir/$file_id/");

  &debug("w_c: start dumping incoming msg into $scandir/$wmaildir/tmp/$file_id [",&deltatime,"]");
  open(TMPFILE,">$scandir/$wmaildir/tmp/$file_id")||&tempfail("cannot write to $scandir/$wmaildir/tmp/$file_id - $!");
  
  my $still_headers=1;
  while (<STDIN>) {
    if ($still_headers) {
      $HEADERS .= $_;
      #Catch any naughty illegal header chars here
      if ($BAD_MIME && /\r|\0/) {
	$illegal_mime=1;
	&debug("w_c: illegal MIME chars found in header");
      }
      #Put headers into array
      if (/^\s+(.*)$/ && $last_hdr) {
	#Hmmm, a continuation...
	$headers{$last_hdr} .= $1;
      } elsif (/^([^\s]+):(.*)$/) {      
	$hdr=$1;
	$last_hdr=tolower($hdr);
	$value=$2;
	$value=~s/^\s//;
	$num_of_headers++;
	#Don't let this array grow without bounds...
	if ($num_of_headers < $MAX_NUM_HDRS) {
	  if ($hdr =~ /^to|cc/i && $headers{tolower($hdr)}) {
	    #Special-case the To: and Cc: headers.
	    #Broken mailers generate messages with multiple 
	    #instances of these, so merge them into one...
	    $headers{tolower($hdr)} .= ",$value";
	  } elsif ($hdr =~ /^(from|x-mail|User-Agent|Organi|Received|Message-ID|Subject)/i && $headers{tolower($hdr)}) {
	    #Make sure any multiples of these headers are remembered, so that 
	    #perlscanner checks can see all instances - just wrap em up
	    #into one long line
	    $headers{tolower($hdr)} .= " $value";
	  } else {
	    #All other headers: the last occurance wins!
	    $headers{tolower($hdr)}=$value;
	  }
	}
      }
      if (/^(\r|\r\n|\n)$/) {
	$still_headers=0;
	#Try to workaround those nasty broken viruses
	if ($BAD_MIME && $headers{'content-type'} ne "" && $headers{'mime-version'} eq "") {
	  print TMPFILE "$V_HEADER-$VERSION: added fake MIME-Version header\nMIME-Version: 1.0\n";
	  $headers{'mime-version'}="1.0";
	  &debug("w_c: added fake MIME-Version header");
	}
	if ($headers{'message-id'} eq "" && !$headers{$qsmsgid}) {
	  $headers{$qsmsgid}="<".time . __LINE__ . $$ . "\@$hostname>";
	  print TMPFILE "$V_HEADER-Message-ID: $headers{$qsmsgid}\n";
	} else {
	  if (!$headers{$qsmsgid}) {
	    $headers{$qsmsgid}=$headers{'message-id'};
	  }
	}
      }
    }
    &check_and_grab_uuencoding;
    print TMPFILE ;
  }
  close(TMPFILE)||&tempfail("cannot close $scandir/$wmaildir/tmp/$file_id - $!");
  &debug("w_c: rename new msg from $scandir/$wmaildir/tmp/$file_id to $scandir/$wmaildir/new/$file_id [",&deltatime,"]");
  rename("$scandir/$wmaildir/tmp/$file_id","$scandir/$wmaildir/new/$file_id")||&tempfail("cannot rename $scandir/$wmaildir/tmp/$file_id into $scandir/$wmaildir/new/$file_id - $!");
  my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$atime,$mtime,$ctime,$blksize,$blocks);
  ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$msg_size,$atime,$mtime,$ctime,$blksize,$blocks) = stat("$scandir/$wmaildir/new/$file_id");
  if (!$headers{'date'}) {
    my (@day, @mon);
    $day[0]='Sun';$day[1]='Mon';$day[2]='Tue';$day[3]='Wed';$day[4]='Thu';$day[5]='Fri';$day[6]='Sat';
    $mon[0]='Jan';$mon[1]='Feb';$mon[2]='Mar';$mon[3]='Apr';$mon[4]='May';$mon[5]='Jun';$mon[6]='Jul';$mon[7]='Aug';$mon[8]='Sep';$mon[9]='Oct';$mon[10]='Nov';$mon[11]='Dec';
    my ($tm_sec,$tm_min,$tm_hour,$tm_mday,$tm_mon,$tm_year,$tm_wday,$tm_yday,$tm_isdst);
    ($tm_sec,$tm_min,$tm_hour,$tm_mday,$tm_mon,$tm_year,$tm_wday,$tm_yday,$tm_isdst)=localtime;
    $tm_year += 1900;
    $headers{'date'}=$day[$tm_wday].", $tm_mday ".$mon[$tm_mon]." $tm_year $tm_hour:$tm_min:$tm_sec";
  }
  #  close(STDIN)||&tempfail("cannot close fd 0 - $!");
}

sub grab_envelope_hdrs {
  select(STDOUT); $|=1;
  
  open(SOUT,"<&STDOUT")||&tempfail("cannot dup fd 0 - $!");
  while (<SOUT>) {
    ($env_returnpath,$env_recips) = split(/\0/,$_,2);
    if ( ($returnpath=$env_returnpath)=~s/^F(.*)$// ) {
      $returnpath=$1;
      ($recips=$env_recips)=~ s/^T//;
      $recips=~ s/\0T/\,/g;
      $recips=~ /^(.*)\0+$/;
      $recips=$1;
      $recips=~ s/\0+$//g;
    }
    #only meant to be one line!
    last;
  }
  close(SOUT)||&tempfail("cannot close fd 1 - $!");
  if ( $env_returnpath eq "" && $env_recips eq "" ) {
    #At the very least this is supposed to be $env_returnpath='F' - so
    #qmail-smtpd must be officially dropping the incoming message for
    #some (valid) reason (including the other end dropping the connection).
    &debug("g_e_h: no sender and no recips.");
    &cleanup;
    exit;
  }
  &debug("g_e_h: return-path is \"$returnpath\", recips is \"$recips\"");
}


sub deconstruct_msg {
  my ($start_decon_time) = [gettimeofday];
  my $save_filename ='';
  my ($new_filename,$MAYBETNEF,$tnef_status);

  &debug("d_m: starting $mimeunpacker_binary <$scandir/$wmaildir/new/$file_id [",&deltatime,"]");
  open(MIME,"$mimeunpacker_binary  <$scandir/$wmaildir/new/$file_id|")||&tempfail("cannot call mimeunpacker - $!");
  while (<MIME>) {
    next if (/exists\./);
    &tempfail("d_m: output spotted from reformime - that shouldn't happen!");
  }
  close(MIME)||&tempfail("cannot close mimeunpacker - $!");
  my $unpacker='';
  
  opendir(DIR,"$scandir/$file_id/")||&tempfail("cannot open dir $scandir/$file_id/ - $!"); 
  my @allfiles = grep(!/^\.+$/, readdir(DIR));
  closedir(DIR);
  &debug("d_m: finished $mimeunpacker_binary [",&deltatime,"]");
  #If you have the tnef app, you'll be able to scan broken M$ attachments
  
  if ( $tnef_binary ) {
    &debug("d_m: Checking all attachments to see if they're MS-TNEF");
    foreach $save_filename (@allfiles) {
      #Clean up $save_filename so as to keep taint happy
      $save_filename=~/^(.*)$/; $save_filename=$1;
      ($new_filename=$save_filename) =~ s/([^a-z0-9\.\-\_\+\=\~]+)//gi;
      if ($save_filename ne $new_filename) {
	$new_filename=~/(\.[^\.]+)$/;
	$new_filename=&uniq_id."$new_filename";
	rename($save_filename,$new_filename);
	&debug("d_m: ren $save_filename to $new_filename");
	$save_filename=$new_filename;
      }
      #Who cares if it is or isn't tnef, just scan it!
      if ($tnef_binary) {
	$MAYBETNEF=`$tnef_binary --number-backups -d $scandir/$file_id/ -f $scandir/$file_id/$save_filename 2>&1`;	
	$tnef_status=$?;
	&debug("d_m: is $scandir/$file_id/$save_filename is a TNEF file?: $tnef_status [",&deltatime,"]");
      }
    }
  }

  #If you're happy with your scanners zip file handling, you can 
  #skip this whole section. McAfee's doesn't support zip for starters!
  
  if ($force_unzip) {
    &debug("d_m: Manually unpack any zip files as some virus scanners don't do zip under Unix!");
    #Re-initialize directory listing
    opendir(DIR,"$scandir/$file_id/")||&tempfail("cannot open dir $scandir/$file_id/ - $!"); 
    @allfiles = grep(!/^\.+$/, readdir(DIR));
    closedir(DIR);
    foreach $save_filename (@allfiles) {
      #Clean up $save_filename so as to keep taint happy
      $save_filename=~/^(.*)$/; $save_filename=$1;
      ($new_filename=$save_filename) =~ s/([^a-z0-9\.\-\_\+\=\~]+)//gi;
      if ($save_filename ne $new_filename) {
	$new_filename=~/(\.[^\.]+)$/;
	$new_filename=&uniq_id."$new_filename";
	rename($save_filename,$new_filename);
	&debug("d_m: ren $save_filename to $new_filename");
	$save_filename=$new_filename;
      }
      if ( $save_filename =~ /\.(zip|exe)$/i) {
	&unzip_file($save_filename);
      }
    }
  }
  if (!$redundant_scanning) {
    if (-f "$scandir/$file_id/$save_filename") {
      system $rm_binary,"-f","$scandir/$file_id/$save_filename";
    }
  }
  my($decon_time)=tv_interval ($start_decon_time, [gettimeofday]);
  &debug("d_m: unpacking message took $decon_time seconds");
}

sub init_scanners {
  my($start_init_scanners_time)=[gettimeofday];
  &debug("ini_sc: start scanning");
  chdir("$scandir/$file_id/");
  #Run perlscan_scanner
  &perlscan_scanner;
  
  #Virus? No point in running the rest then
  return if ($quarantine_event);

  #Delete original zip'ped attachment as there's no point 
  #in the other scanners double-scanning it - unless $redundant scanning
  #is set....
  if ($redundant_scanning) {
    link("$scandir/$wmaildir/new/$file_id","$scandir/$file_id/orig-$file_id");
  }
  &debug("ini_sc: recursively scan the directory $scandir/$file_id/");
  &scanloop;
  chdir("$scandir");
  my($decon_time)=tv_interval ($start_init_scanners_time, [gettimeofday]);
  &debug("ini_sc: scanning message took $decon_time seconds");
}


sub perlscan_scanner {
  #This is most efficient if called from within deconstruct_msg

  my($start_perlscan_time)=[gettimeofday];
  my (%array,$var,$lfile,$filename,$section,$apptype,$save_filename);
  my ($type,$desc,$file,@allfiles);
  my ($ps_skipfile,$extension);
  my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks,$fsize);
  my ($attachment_list,$perlscan_time);
  &debug("p_s: starting scan of directory \"$scandir/$file_id\"...");
  
  use DB_File;
  
  
  tie (%array, 'DB_File', "$db_filename.db", O_RDONLY, 0600) || &tempfail("cannot open  $db_filename - $!");

  if ($illegal_mime && $headers{'mime-version'} && $BAD_MIME) {
    $destring="problem";
    $quarantine_description="Disallowed MIME characters found in headers";
    $quarantine_event="Policy:Bad_MIME";
    $description .= "\n---perlscanner results ---\n$destring '$quarantine_description'\n found in message $scandir/$file_id";
  }
  #check out headers against DB...
  
  foreach $var (sort keys(%array)) {
    ($type,$desc)=split(/\t/,$array{$var},2);
    &debug("p_s:  '$var' = '$type' = '$desc'");
    if ($type !~ /^[0-9]+$/) {
      &debug("p_s:  type is a header!");
      #Strip off numeric uid...
      $var=~ s/^[0-9]+://;
      $type =~ s/^Virus-//g;
      &debug("p_s:  checking for objects containing $type: $var");
      if ($headers{$type} =~ /^$var$/) {
	$quarantine_description="$desc";
	($quarantine_event=$quarantine_description)=~s/\s/_/g;
	$quarantine_event="Perlscan:".substr($quarantine_event,0,$QE_LEN);
	$description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in file $scandir/$file_id/$file";
	&debug("p_s: something to block! ($quarantine_description)");
      }
    } else {
      &debug("p_s: type is a size!");
    }
  }

  opendir(DIR,"$scandir/$file_id/")||&tempfail("cannot open dir $scandir/$file_id/ - $!"); 
  @allfiles = grep(!/^\.+$/, readdir(DIR));
  closedir(DIR);

  #append any ORIGINAL uuencoded filenames to this directory array
  #so that perlscanner can match on uuencoded filenames
  foreach $file (@allfiles,@uufile_list,@zipfile_list) {
    #skip files that reformime generates.
    #This will potentially allow baddies to smuggle files through
    #by using filenames like this... Nothing can be done about that:-(
    #Reformime generates filenames of the form:
    # 967067231.24320-X.host.name (where X is a number)
    if ($file =~ /^[0-9]+\.[0-9]+\-[0-9]+\.|^$file_id/) {
      &debug("p_s: skipping auto-generated file $file");
      $ps_skipfile=1;
    } else {
      &debug("p_s: checking $file against perlscanner database...");
      $ps_skipfile=0;
    }

    # Grab extension for wildcard matches
    if ($file=~/(^.*)(\.[^\.]+)$/) {
      $extension=tolower($2);
    } else {
      $extension="";
    }
    $lfile = tolower($file);
    &debug("p_s: file $file is lowercased to $lfile and has extension $extension") if (!$ps_skipfile);
    ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = stat("$file");
    if ($size) {
      #Sanity check so that the virtual attachments don't get double-counted
      $file_desc .= "$file:$size\t";
    }
    &debug("p_s: compare $lfile against perlscanner database") if (!$ps_skipfile);
    if ( ($array{$lfile} || $array{$extension}) && !$ps_skipfile ) {
      if ($array{$lfile}) {
	($fsize,$quarantine_description) = split(/\t/,$array{$lfile},2);
      } else {
	$destring="Disallowed attachment type";
	($fsize,$quarantine_description) = split(/\t/,$array{$extension},2);
      }
      $attachment_list.="$file:$size,";
      if ($size eq $fsize || $fsize =~ /^(\-|\*|any|0)$/i ) {
	&debug("p_s: Quarantine $file! ($quarantine_description)");
	($quarantine_event=$quarantine_description)=~s/\s/_/g;
	$quarantine_event="Perlscan:".substr($quarantine_event,0,$QE_LEN);
	$description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in file $scandir/$file_id/$file";
	$section=$apptype=$save_filename=$filename="";
#	return;
      }
    }
  }
  untie %array;
  chdir("$scandir/");
  my($stop_perlscan_time)=[gettimeofday];
  $perlscan_time = tv_interval ($start_perlscan_time, $stop_perlscan_time);
  &debug("p_s:  finished scan of dir \"$scandir/$file_id\" in $perlscan_time secs");
}


sub scanloop {
  &debug("scanloop: starting scan of directory \"$scandir/$file_id\"...");

  my ($scanner);

  foreach $scanner (@scanner_array) {
    #Any scanner errors caused by broken zip files/etc will be ignored
    # - not sure how that should be handled...
    &{$scanner};
    if ($quarantine_event) {
      #If one scanner finds a virus - why run the rest over it?
      last;
    }
  }
  &debug("scanloop: finished scan of \"$scandir/$file_id\"...");
}

sub qmail_requeue {
  my($sender,$env_recips,$msg)=@_;
  my ($temp,$findate);

  &debug("q_r: fork off child into $qmailqueue...");

  ($recips=$env_recips)=~ s/^T//;
  $recips=~ s/\0T/\,/g;
  $recips=~ /^(.*)\0+$/;
  $recips=$1;
  $recips=~ s/\0+$//g;
  
  # Create a pipe through which to send the envelope addresses.
  pipe (EOUT, EIN) or &tempfail("Unable to create a pipe. - $!");
  select(EOUT);$|=1;
  select(EIN);$|=1;
  # Fork qmail-queue.  The qmail-queue child will then open fd 0 as
  # $message and fd 1 as the reading end of the envelope pipe and exec
  # qmail-queue.  The parent will read in the addresses and pass them 
  # through the pipe and then check the exit status.

  $elapsed_time = tv_interval ($start_time, [gettimeofday]);
  local $SIG{PIPE} = 'IGNORE';
  my $pid = fork;

  if (not defined $pid) {
    &tempfail ("Unable to fork. (#4.3.0) - $!");
  } elsif ($pid == 0) {
    # In child.  Mutilate our file handles.
    close EIN; 
    
    open(STDIN,"<$msg")|| &tempfail ("Unable to reopen fd 0. (#4.3.0) - $!");

    open (STDOUT, "<&EOUT") ||  &tempfail ("Unable to reopen fd 1. (#4.3.0) - $!");
    select(STDIN);$|=1;
    open(QMQ,"|$qmailqueue")||&tempfail("Unable to fork $qmailqueue (#4.3.0) - $!");
    ($sec,$min,$hour,$mday,$mon,$year) = gmtime(time);
    $elapsed_time = tv_interval ($start_time, [gettimeofday]);
    $findate = POSIX::strftime( "%d %b ",$sec,$min,$hour,$mday,$mon,$year);
    $findate .= sprintf "%02d %02d:%02d:%02d -0000", $year+1900, $hour, $min, $sec;
    print QMQ "Received: from $returnpath by $hostname by uid $uid with qmail-scanner-$VERSION \n";
    print QMQ " ($SCANINFO Clear:$tag_score. \n";
    print QMQ " Processed in $elapsed_time secs); $findate\n";
    print QMQ "X-Spam-Status: $sa_comment\n" if ($sa_comment ne ""); 
    if ( $descriptive_hdrs ) {
      print QMQ "${V_HEADER}-Mail-From: $returnpath via $hostname\n";
      print QMQ "${V_HEADER}-Rcpt-To: $recips\n" if ($descriptive_hdrs eq "2");
      print QMQ "$V_HEADER: $VERSION (Clear:$tag_score. Processed in $elapsed_time secs)\n";
    }
    my $still_headers=1;
    while (<STDIN>) {
      if ($still_headers) {
	next if ($sa_comment ne "" && /^X-Spam-Status:/i);
	if ($sa_comment =~ /^yes/i && /^(Subject):(\s?)([^\n]+)\n/i && $spamc_subject ne "" ) {
	  $altered_subject="$1: $spamc_subject $3";
	  if ($altered_subject !~ /^: $spamc_subject/) {
	    &debug("altering subject line to $altered_subject");
	    print QMQ "$altered_subject\n";
	    next;
	  }
	}
	$still_headers=0 if (/^(\r|\r\n|\n)$/);
      }
      print QMQ;
    }
    close(QMQ)||&tempfail("Unable to close pipe to $qmailqueue (#4.3.0) - $!");
    #This child is finished - exit
    exit;
  } else {
    # In parent.
    close EOUT;
      
    # Feed the envelope addresses to qmail-queue.
    print EIN "$sender\0$env_recips";
    close EIN  || &tempfail ("Write error to envelope pipe. (#4.3.0) - $!");
}

  # We should now have queued the message.  Let's find out the exit status
  # of qmail-queue.
  waitpid ($pid, 0);
  my $status =($? >> 8);
  if ($status != 0) { &tempfail ("Unable to queue message ($status). (#4.3.0) - $!" )}
  
}


sub valid_virus_to_report {
  my ($virus_type)=@_;
  my ($virus)='';
  # This subroutine is used to determine if the virus found during the scan
  # is reportable. i.e. do we want to send a message to this user or not as is
  # the case with the KLEZ virus.
  &debug("v_v_t_r: called with $virus_type");
  foreach $virus (@silent_viruses_array) {
    &debug("v_v_t_r: does $virus_type contain $virus?");
    if ($virus_type =~ /$virus/i) {
      &debug("v_v_t_r: yes it does! - so don't notify the sender");
      return 0;
    }
  }
  return 1;
}

sub is_unreplyable_email {
  my ($addr_type)=@_;
  my ($dom,$is_local)='';
  #This subroutine is used to see if the sender of this message
  #was a mailing-list/postmaster/etc, or the recipient is a local user. 
  #If it is we don't want to send a reply.
  &debug("i_u_e: called with $addr_type");
  
  if ($addr_type eq "recips") {
    foreach $dom (@local_domains_array) {
      &debug("i_u_e: does $recips contain $dom?");
      if ($recips =~ /$dom$/i) {
	&debug("i_u_e: yes it does!");
	$is_local++;
      }
    }
  } else {
    $is_local="99";
    if ($headers{'x-loop'} || $headers{'x-listname'} || $headers{'x-listmember'} || $headers{'mailing-list'} || $headers{'x-mailing-list'} || $headers{'precedence'} =~ /^(bulk|list|junk)$/i || $returnpath =~ /^$|daemon|request|bounce|mailer|postm|owner|lists|words|majordom|experts|\-(return|error)/i ) {
      &debug("i_u_e: $addr_type is a mailing-list");
      return 1;
    }
  }
  #
  #Only reply if it is a local address
  if (!$is_local) {
    &debug("i_u_e: is_local=$is_local");
    return 1;
  } else {
    &debug("i_u_e: is_local=$is_local");
    return 0;
  }
}

sub email_quarantine_report {
  my($start_email_time)=[gettimeofday];
  my($one_recip);
  &debug("e_v_r: quarantine  msg to $scandir/$vmaildir/new/$file_id");
  rename("$scandir/$wmaildir/new/$file_id","$scandir/$vmaildir/new/$file_id")||&tempfail("cannot rename $scandir/$wmaildir/new/$file_id into $scandir/$vmaildir/new/ - $!");
  open(QTINE,">>$scandir/$vmaildir/new/$file_id");
  print QTINE "\n*** Qmail-Scanner Quarantine Envelope Details Begin ***\n";
  print QTINE "${V_HEADER}-Mail-From: \"$returnpath\" via $hostname\n";
  print QTINE "${V_HEADER}-Rcpt-To: \"$recips\"\n";
  print QTINE "$V_HEADER: $VERSION ($SCANINFO $destring Found. Processed in ",tv_interval($start_time,[gettimeofday])," secs)\n";
  print QTINE "*** Qmail-Scanner Envelope Details End ***\n";
  close QTINE;
  &email_sender("sender") if (&valid_virus_to_report($quarantine_description));
  &email_sender("admin");
  if ($recips =~ /\,/) {
    for $one_recip (split(/\,/,$recips)) {
      &email_recips($one_recip);
    }
  } else {
    &email_recips($recips);
  }
  &write_quarantine_report;
  $elapsed_time = tv_interval ($start_time, [gettimeofday]);
  &debug("e_v_r: email_quarantine_report took ".tv_interval ($start_email_time, [gettimeofday])." seconds to execute");
}

sub cleanup {
  chdir("$scandir/");
  if (!$archiveit) {
    &debug("cleanup: $rm_binary -rf $scandir/$file_id/ $scandir/$wmaildir/new/$file_id") ;
  } else {
    if ( -f "$scandir/$wmaildir/new/$file_id") {
      &debug("cleanup: archiving into $scandir/$archivedir/new/");
      rename("$scandir/$wmaildir/new/$file_id","$scandir/$archivedir/new/$file_id");
      #This will do for now. Not pretty - but very cheap!
      open(ARCHIVE,">>$scandir/$archivedir/new/$file_id");
      print ARCHIVE "\n*** Qmail-Scanner Envelope Details Begin ***\n";
      print ARCHIVE "${V_HEADER}-Mail-From: \"$returnpath\" via $hostname\n";
      print ARCHIVE "${V_HEADER}-Rcpt-To: \"$recips\"\n";
      print ARCHIVE "$V_HEADER: $VERSION ($SCANINFO  Clear:$tag_score. Processed in ",tv_interval($start_time,[gettimeofday])," secs)\n";
      print ARCHIVE "*** Qmail-Scanner Envelope Details End ***\n";
      close ARCHIVE;
    }
  }
  system("$rm_binary -rf $scandir/$file_id/ $scandir/$wmaildir/new/$file_id") if ($DEBUG < 100);

}


sub scan_queue {
  my ($scanner,$SCANINFO,$files,$sweep_eng,$sweep_product,$dir);
  my $start_scan_time =time;
  my ($inocucmd_eng,$inocucmd_product,$spamassassin_eng);
  
  chdir($scandir);
  &debug("s_q: re-create the quarantine version file");
  foreach $scanner (@scanner_array) {
    $scanner=~s/_scanner//;
    &debug("s_q: detecting version of $scanner");
    if ($scanner eq "uvscan") {
      open(UV,"$uvscan_binary --version|")||die "failed to call $uvscan_binary --version - $!";
      while (<UV>) {
	chomp;
	if (/^Scan engine (v[0-9\.]+) /) {
	  $SCANINFO .="uvscan: $1/";
	} elsif (/^Virus data file (v[0-9\.]+) /) {
	  $SCANINFO .= "$1. ";
	}
      }
      close(UV);
    } elsif ($scanner eq "csav") {
      open(CS,"$csav_binary -virno|")||die "failed to call $csav_binary -virno - $!";
      while (<CS>) {
	chomp;
	if (/Command Software AntiVirus for Linux (version [0-9\.]+)/) {
	  $SCANINFO .="csav: $1/";
	} elsif (/^CSA[V]: (.*)/) {
	  $SCANINFO .= "$1/";
	}
      }
      close(CS);
    } elsif ($scanner eq "trophie" ) {
      open(IS,"$trophie_binary -v 2>&1|")||die "failed to call $trophie_binary -v - $!";
      while (<IS>) {
	chomp;
	if (/VSAPI version (.*)/) {
	  $SCANINFO .= "trophie: $1/";
	} elsif (/Pattern version ([0-9]+) \(pattern number ([0-9]+)\)/) {
	  $SCANINFO .= "$1/$2. ";
	}
      }
      close(IS);
    } elsif ($scanner eq "iscan") {
      open(IS,"$iscan_binary -v|")||die "failed to call $iscan_binary -v - $!";
      while (<IS>) {
	chomp;
	if (/Virus Scanner (v[0-9\.]+), VSAPI (v[0-9\.\-]+)/) {
	  $SCANINFO .="iscan: $1/$2/";
	} elsif (/Pattern version ([0-9\.]+)/) {
	  $SCANINFO .= "$1/";
	} elsif (/Pattern number ([0-9\.]+)/) {
	  $SCANINFO .= "$1. ";
	}
      }
      close(IS);
    } elsif ($scanner eq "fsecure") {
      open(FS,"$fsecure_binary --version|")||die "failed to call $fsecure_binary --version - $!";
      while (<FS>) {
        chomp;
        if (/Release ([0-9\.]+) build ([0-9]+)/) {
          $SCANINFO .="fsecure: $1/$2/";
        } elsif (/sign.def version ([0-9\.]+-[0-9\.]+-[0-9\.]+)/) {
          $SCANINFO .= "$1/";
        } elsif (/fsmacro.def version ([0-9\.]+-[0-9\.]+-[0-9\.]+)/) {
          $SCANINFO .= "$1/";
        } elsif (/sign2.def version ([0-9\.]+-[0-9\.]+-[0-9\.]+)/) {
          $SCANINFO .= "$1. ";
        }
      }
      close(FS);
    } elsif ($scanner eq "fprot") { 
      open(FP,"$fprot_binary \?|")||die "failed to call $fprot_binary --version - $!";
      while (<FP>) {
	chomp;
	if (/F-PROT ([0-9\.]+)/) {
	  $SCANINFO .="f-prot: $1";
	} elsif (/SIGN.DEF version ([0-9\.]+-[0-9\.]+-[0-9\.]+)/) {
	  $SCANINFO .= "$1/";
	} elsif (/MACRO.DEF version ([0-9\.]+-[0-9\.]+-[0-9\.]+)/) {
	  $SCANINFO .= "$1/";
	} elsif (/SIGN2.DEF version ([0-9\.]+-[0-9\.]+-[0-9\.]+)/) {
	  $SCANINFO .= "$1";
	}
      }
      $SCANINFO .= ". ";
      close(FP);
    } elsif ($scanner eq "hbedv") {
      open(IS,"$hbedv_binary --version 2>&1 |")||die "failed to call  $hbedv_binary --version - $!";
      while (<IS>) {
	chomp;
	$SCANINFO .= "hbedv: $_. ";
      }
      close(IS);
    } elsif ($scanner eq "avp") {
      open(AVP,"$avp_binary -Y -VL 2>&1 |")||die "failed to call  $avp_binary -Y -VL  - $!";
      while (<AVP>) {
	chomp;
	if (/Version ([0-9\.]+)\s+build ([0-9\.]+)/) {
	  $SCANINFO .= "avp: $1/$2. ";
	}
      }
      close(AVP);
    } elsif ($scanner eq "ravlin") {
      open(RAV,"$ravlin_binary --version 2>&1 |")||die "failed to call  $ravlin_binary --version  - $!";
      while (<RAV>) {
	chomp;
	if (/^Version: ([0-9\.]+)\./) {
	  $SCANINFO .= "ravlin: $1. ";
	}
      }
      close(RAV);
    } elsif ($scanner eq "sophie") {
      open(SOP,"$sophie_binary -v 2>&1|")||die "failed to call $sophie_binary -v - $!";
      while (<SOP>) {
	chomp;
	if (/Sophos engine version (.*)$/) {
	  $sweep_eng=$1;
	} elsif (/Sophos IDE version ([0-9\.]+)/) {
	  $sweep_product=$1;
	}
      }
      $SCANINFO .= "sophie: $sweep_eng/$sweep_product. ";
      close(SOP);
    } elsif ($scanner eq "sweep") {
      open(SOP,"$sweep_binary -v|")||die "failed to call $sweep_binary -v - $!";
      while (<SOP>) {
	chomp;
	if (/Engine version\s+:\s+(.*)$/) {
	  $sweep_eng=$1;
	} elsif (/Product version\s+:\s+(.*)$/) {
	  $sweep_product=$1;
	}
      }
      $SCANINFO .= "sweep: $sweep_eng/$sweep_product. "; 
      close(SOP);
    } elsif ($scanner eq "inocucmd") {
      open(IOP,"$inocucmd_binary -HEL|")||die "failed to call $inocucmd_binary -HEL - $!";
      while (<IOP>) {
	chomp;
	if (/Engine version:\s+(.*) ([0-9\/]+)$/) {
	  $inocucmd_eng=$1;
	} elsif (/Data version:\s+(.*) ([0-9\/]+)$/) {
	  $inocucmd_product=$1;
	}
      }
      $SCANINFO .= "inocucmd: $inocucmd_eng/$inocucmd_product. ";
      close(IOP);
    } elsif ($scanner eq "clamscan") {
      open(CLAMS,"$clamscan_binary --stdout -V|")||die "failed to call $clamscan_binary --stdout -V - $!";
      while (<CLAMS>) {
	chomp;
	if (/ersion ([0-9\.]+)/) {
	  $SCANINFO .="clamscan: $1. ";
	}
      }
      close(CLAMS);
    } elsif ($scanner eq "spamassassin") {
      #X-Spam-Checker-Version: SpamAssassin 2.01
      open(SPAS,"$spamassassin_binary  -V |")||die "failed to call $spamassassin_binary -V - $!";
      $spamassassin_eng="2.x";
      while (<SPAS>) {
	chomp;
	if (/^SpamAssassin version (.*)$/i) {
	  $spamassassin_eng=$1;
	}
      }
      close(SPAS);
      $SCANINFO .= "spamassassin: $spamassassin_eng. ";
    } else {
      #Catch-all for other ones
      $SCANINFO .= "$scanner: ???. ";
    }
  }
  open(VER,">$versionfile")||die "cannot write to $versionfile - $!";
  print VER $SCANINFO;
  close(VER);
  
  &debug("s_q: cleaning up old files via $find_binary $scandir -type f \! -name '*.log' \! -name '*.txt' \! -name '*.db' \! -path '*/quarantine/*' -prune  -mmin +2160 -exec $rm_binary -f {} \;");
  my ($OLDFILES)=`$find_binary $scandir -type f \! -name '*.log' \! -name '*.txt' \! -name '*.db' \! -path '*/quarantine/*' -prune  -mmin +2160 -exec $rm_binary -f {} \\; 2>/dev/null`;
}

sub write_quarantine_report {
  my ($temp,$desc,$report,$subj);
  $subj=$headers{'subject'};
  $subj=~s/\t/ /g;
  $desc=$quarantine_description;
  $desc=~s/\n\t/ /g;
  $report = "$nowtime\t$returnpath\t$recips\t$subj\t$desc\t$SCANINFO\n";
  open(QUARANTINELOG,">>$scandir/$quarantinelog");
  print QUARANTINELOG $report;
  close QUARANTINELOG;
  &debug("w_v_r: writing quarantine log report of: $report");
}

sub scanner_info {
  open(SC,"<$versionfile")||&tempfail("cannot open $versionfile - did you initialise the system by running \"$prog -z\"? - $!");
  $SCANINFO = <SC>;
  $SCANINFO =~ s/\n|\r|\0/ /g;
  close(SC);
}

sub generate_quarantine_db {
  use DB_File;
  use vars qw( %h);
  my ($line,%array,$count,$match,$type,$descr,$entry,$descrip,$size);
  if ($opt_g) {
    print "perlscanner: generate new DB file\n";
   unlink("$db_filename.db.tmp");
    tie (%array, 'DB_File', "$db_filename.db.tmp", O_CREAT|O_RDWR, 0640, $DB_HASH ) || &tempfail("cannot open for write $db_filename.db.tmp - $!");
    
    open(TXT,"<$db_filename.txt")||&tempfail("cannot read $db_filename.txt - $!");

    #Remeber: all filenames are lowercased, but headers aren't...
    while (<TXT>) {
      $line++;
      next if (/^\#|^\s+$/); #ignore lines starting with hashes
      chomp;
      $count++;
      ($match,$type,$descr)=split(/\t+/,$_,3);
      if ( $match eq "" || ($type !~ /^[0-9]+$/ && $type !~ /^Virus-[0-9a-z\_\-]+:$/i) ) {
	print "ERROR: incorrect format on line $line - skipped\n";
      } else {
	#Strip off any regex endings
	if ($type =~ /^[0-9]+$/) {
	  #this is a filename/attachment
	  if ( $match =~ /^\.dat$/i ) {
	    print "ERROR: on line $line. Cannot block all .dat files. Will block too many normal messages. Skipped...\n";
	    next;
	  }
	  $match = tolower($match);
	} else {
	  #this is for header matches
	  $match =~ s/^\^|\$$//g;
	  #Now make unique
	  $match = "$line:$match";
	  $type=~s/:$//;
	  $type =~ /^Virus\-(.*)/;
	  if ($1 !~ /^(MAILFROM|RCPTTO|TCPREMOTEIP)$/) {
	    $type="Virus-".tolower($1);
	  }
	}
	$array{"$match"}="$type\t$descr";
      }
    }
    close(TXT);
#    $array->sync;
    untie %array ;
    rename("$db_filename.db.tmp","$db_filename.db");
    print "perlscannner: total of $count entries.\n";
  } else {
    print "perlscanner: reading from $db_filename.db\n";
    tie (%array, 'DB_File', "$db_filename.db", O_RDONLY, 0600) || &tempfail("cannot open  $db_filename - $!");
    foreach $entry (keys %array) {
      $count++;
      ($type,$descrip)=split(/\t/,$array{$entry},2);
      if ( $type =~ /^([0-9]+|Any)/) {
	if ($type eq "0") {
	  $type="Any";
	} elsif ($size =~ /^[0-9]+$/) {
	  $type="$type bytes";
	}
	print "File: \t$entry\n\t\t\tSize: $type\n\t\t\tDescription: $descrip\n\n";
      }
      if ($type =~ /^Virus-(.*)$/i) {
	$type=$1;
	#Strip off numeric uid...
	$entry=~ s/^[0-9]+://;
	if ($type =~ /^(MAILFROM|RCPTTO|TCPREMOTEIP)$/) {
	  print "Envelope Header: \t$type\n\t\t\tContent: ^$entry\$\n\t\t\tDescription: $descrip\n\n";
	} else {
	  print "Email Header: \t$type\n\t\t\tContent: ^$entry\$\n\t\t\tDescription: $descrip\n\n";
	}
      }
    }
    untie %array;
    print "perlscanner: total of $count entries found.\n";
  }
}




sub show_version {
  my ($scanner);
  &scanner_info;
  print "

$prog

Version: $VERSION

Perl:    Summary of my perl5 (revision 5.0 version 6 subversion 1) configuration:
    /usr/lib/perl5/5.6.1/i386-linux
    /usr/lib/perl5/5.6.1
    /usr/lib/perl5/site_perl/5.6.1/i386-linux
    /usr/lib/perl5/site_perl/5.6.1
    /usr/lib/perl5/site_perl
    /usr/lib/perl5/vendor_perl/5.6.1/i386-linux
    /usr/lib/perl5/vendor_perl/5.6.1
    /usr/lib/perl5/vendor_perl

Scanners: perlscanner";
  foreach $scanner (@scanner_array) {
    print ", $scanner";
  }

  print "\n\nScanner versioning: $SCANINFO\n";
  print "
Operating System: Linux, 2.4.18-5smp
Hardware:         i686";
  print "\n\n\n";
}


sub email_sender {
  my($addr_type)=@_;
  my ($HDR,$hdr,$tmpsndrs,$tmpsubj)='';
  my ($tmpmsgid)= &uniq_id() . "-" . $V_FROM;
  open(SM,"|$qmailinject -h -f ''")||&tempfail("cannot open $qmailinject for sending quarantine report - $!");
  print SM "From: \"$V_FROMNAME\" <$V_FROM>\n";
  if ($addr_type eq "sender") {
    if (!&is_unreplyable_email('sender') && &notify_addr('sender')) {
      &debug("e_s: sending quarantine report via: $qmailinject to sender address ($returnpath)");
      print SM "To: $returnpath\n";
      $tmpsndrs = "$returnpath";
    } else {
      return;
    }
  } else {
    if (&notify_addr('admin') || (&notify_addr('nmladm') && !&is_unreplyable_email('sender'))) {
      &debug("e_s: sending quarantine report via: $qmailinject to admin address ($QUARANTINE_CC)");
      print SM "To: $QUARANTINE_CC\n";
      $tmpsndrs .= "$QUARANTINE_CC";
    }
  }
  $tmpsubj="$destring found in sent message \"$headers{'subject'}\"";
  $tmpsubj =~ s/(\r|\0|\n)/ /g;
  print SM "Subject: $tmpsubj\n";
  print SM "Message-ID: <".&uniq_id."\@$hostname>\n";
  print SM "X-Tnz-Problem-Type: 40\n";
  print SM "MIME-Version: 1.0\n";
  print SM "Content-type: text/plain\n";
  if ( $descriptive_hdrs ) {
    print SM "${V_HEADER}-Mail-From: $returnpath via $hostname\n";
    print SM "${V_HEADER}-Rcpt-To: $recips\n" if ($descriptive_hdrs eq "2");
    print SM "$V_HEADER: $VERSION ($SCANINFO $destring Found. \n";
    print SM " Processed in ",tv_interval($start_time,[gettimeofday])," secs)\n";
  }
  print SM "\n";
  if (!&is_unreplyable_email('sender')) {
    print SM "
Attention: $returnpath\n";
    print SM "\n
A $destring was found in an Email message you sent. 
This Email scanner intercepted it and stopped the entire message
reaching its destination. 

The $destring was reported to be: 

$quarantine_description\n";
  } else {
    print SM "
Attention: $V_FROMNAME.\n";
    print SM "
[This warning message is *not* being sent to the apparent originator
of the original message.  This address appears to be that of a
mailing list or other automated email system.]\n";
  }
  if ($destring eq "virus") {
    print SM "\n
Please update your virus scanner or contact your IT support 
personnel as soon as possible as you have a virus on your system.\n";
  } else {
    print SM "\n
Please contact your I.T support personnel with any queries regarding this 
policy.\n";
  }
  print SM "\n
Your message was sent with the following envelope:

MAIL FROM: $returnpath
RCPT TO:   $recips 

... and with the following headers:\n\n";
  print SM "---\n";
  print SM "MAILFROM: $headers{'MAILFROM'}\n";
  print SM "$HEADERS\n";
  print SM "---\n";
  if ($addr_type ne "sender" ) {
    print SM "\n

The original message is kept in:

  $hostname:$scandir/$vmaildir

where the $V_FROMNAME can further diagnose it.

The Email scanner reported the following when it scanned that message:

--- 
$description
---\n";
  }
  close(SM);
  if ($log_details) {
    &log_msg("qmail-scanner","Clear:$tag_score",$elapsed_time,1100,$V_FROM,$tmpsndrs,$tmpsubj,$tmpmsgid,"quarantine-event.txt:1000");
  }
}

sub email_recips {
  my($one_recip)=@_;
  return if (!&notify_addr('recips'));
  my($HDR,$hdr,$tmprecips,$tmpsubj)='';
  my($tmpmsgid)= &uniq_id() . "-" . $V_FROM;

  open(SM,"|$qmailinject -h -f ''")||&tempfail("cannot open $qmailinject for sending quarantine report - $!");
  print SM "From: \"$V_FROMNAME\" <$V_FROM>\n";
  if (!&is_unreplyable_email('recips')) {
    &debug("e_r: sending quarantine report via: $qmailinject to recip address ($one_recip)");
    print SM "To: $one_recip\n";
  }
  $tmpsubj= "$destring found in received message \"$headers{'subject'}\"";
  $tmpsubj =~ s/(\r|\0|\n)/ /g;
  print SM "Subject: $tmpsubj\n";
  print SM "Message-ID: <".&uniq_id."\@$hostname>\n";
  print SM "X-Tnz-Problem-Type: 40\n";
  print SM "MIME-Version: 1.0\n";
  print SM "Content-type: text/plain\n";
  if ( $descriptive_hdrs ) {
    print SM "${V_HEADER}-Mail-From: $returnpath via $hostname\n";
    print SM "${V_HEADER}-Rcpt-To: $one_recip\n" if ($descriptive_hdrs eq "2");
    print SM "$V_HEADER: $VERSION ($SCANINFO $destring Found. \n";
    print SM " Processed in ",tv_interval($start_time,[gettimeofday])," secs)\n";
  }
  print SM "\n";
  print SM "
Attention: $one_recip\n";
  if (!&is_unreplyable_email('recips')) {
    if (&notify_addr('sender')) {
      print SM "
[A message has been sent to the originator, stating there is a virus
in the Email they just sent to you. No further action is required on
your part.]\n";
    }
  } else {
    print SM "

[This message was _not_ sent to the originator address, as that appears to
be a mailing-list or some other automated Email message]\n";
  }
  print SM "\nA $destring was found in an Email message sent to you. 
This Email scanner intercepted it and stopped the entire message
before it reached you. No further action is required on your part.\n";
  print SM "\nThe $destring was reported to be: 

$quarantine_description

Please contact your I.T support personnel with any queries regarding this 
policy.

The message sent to you had the following envelope:

MAIL FROM: $returnpath
RCPT TO:   $recips 

... and with the following headers:\n\n";
  print SM "---\n";
  print SM "MAILFROM: $headers{'MAILFROM'}\n";
  print SM "$HEADERS\n";
  print SM "---\n";
  #print SM "\nLxOCALE_recips_quarantine\n";
  close(SM);
  if ($log_details) {
    &log_msg("qmail-scanner","Clear:$tag_score",$elapsed_time,1100,$V_FROM,$one_recip,$tmpsubj,$tmpmsgid,"quarantine-event.txt:1000");
  }
}

sub notify_addr {
  my($addr_type)=@_;
  &debug("n_a: notify_addr (set to $NOTIFY_ADDRS) called with $addr_type");
  if (($NOTIFY_ADDRS =~ /$addr_type/ || $NOTIFY_ADDRS =~ /all/) && ($NOTIFY_ADDRS !~ /none/)) {
    return 1;
  } else {
    return 0;
  }
}

sub unzip_file {
  my($zipfile)=@_;
  my ($MAYBEZIP,$zfile,$zline,$zsize,$zip_status,$passwd_protected_zip);

  &debug("d_m: potential zip archive file found ($zipfile).");
  &debug ("d_m: it is possibly a zip file, run unzip $unzip_options -t $scandir/$file_id/$zipfile");
  $MAYBEZIP=`$unzip_binary $unzip_options -t $scandir/$file_id/$zipfile 2>&1`;
  $zip_status=($? >> 8);
  
  if ( ($zip_status > 0) && ($zip_status !~ /^(1|2|3|51|81|82)$/) && ($MAYBEZIP !~ /skipping: /) ) {
    &debug("d_m: not a recognisable zip file ($MAYBEZIP)");
  } else {
    &debug ("d_m: it is a zip file - unpack it!");
    &debug("d_m: run $unzip_binary $unzip_options -oj $scandir/$file_id/$zipfile 2>&1");
    open(ZIPPED,"$unzip_binary $unzip_options -oj  $scandir/$file_id/$zipfile 2>&1|")||&tempfail("u_f: cannot open $scandir/$file_id/$zipfile - $!");
    while (<ZIPPED>) {
      if (/^\s+skipping:\s(.*)\s+(encrypted|incorrect password)/) {
	$passwd_protected_zip++;
	#grab these protected filenames for reports anyway.
	$zfile=$1;
	$zfile=~s/^.*\///g;
	$zfile=~s/(^\s+|\s+$)//g;
	
	#$file_desc .= "$zfile:$zsize\t";
      }
    }
    close(ZIPPED);
    $zip_status=($? >> 8);
    if ($zip_status > 0 && ($zip_status !~ /^(1|2|3|51|81|82)$/ && !$passwd_protected_zip)) {
      &tempfail("u_f: cannot close unzip (error code: $zip_status,$passwd_protected_zip) - $!");
    }
    #Only delete original zip file if it happily unpacked.
    if ( $zip_status eq 0 && -f "$scandir/$file_id/$zipfile") {
      system $rm_binary,"-f","$scandir/$file_id/$zipfile";
      &debug("u_f: $zip_status, and successfully unzipped");
      #It may have been deleted, but you still want to see if 
      #it matches the perlscanner DB...
      #push(@zipfile_list, $zipfile);
      my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$zsize,$atime,$mtime,$ctime,$blksize,$blocks);
      ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$zsize,$atime,$mtime,$ctime,$blksize,$blocks) = stat("$zipfile");
      $file_desc .= "$zipfile:$zsize\t";
    } 
  }
}

sub deltatime {
  my ($delta,$current_time,$last_time);
  $current_time = [gettimeofday];
  $delta =  tv_interval ($last_time, $current_time);
  $last_time=$current_time;
  return $delta;
}

sub qmail_parent_check {
  my $ppid=getppid;
  #&debug("q_s_c: PPID=$ppid");
  if ($ppid == 1)  {
    &debug("q_s_c: Whoa! parent process is dead! (ppid=$ppid) Better die too...");
    close(LOG);
    &cleanup;
    #Exit with temp error anyway - just to be real anal...
    exit 111; 
  }
}

sub check_and_grab_uuencoding {
  my ($begin,$perms,$uufile,$uuextension,$uulength,$uuencoded_attachments);
  if (/^(begin) ([0-9][0-9][0-9]) (.*)\n$/) {
    &debug("Ooohhhh, a uuencoded attachment!");
    $uuencoded_attachments++;
    $begin=$1;
    $perms=$2;
    $uufile=$3;
    push(@uufile_list, $uufile);
    $uufile=~s/[^0-9a-z\_\-\.]/_/gi;
    $uufile =~ /\.([^\.]+)$/;
    $uuextension=$1;
    #Ensure the file extension isn't too long either...
    $uuextension=substr($uuextension,0,20); 
    $uulength=length($uufile);
    #Ensure the filelength isn't too large!
    if ( $uulength > $MAX_FILE_LENGTH) {
      &debug("uudecode output: gah! filename is > $MAX_FILE_LENGTH (actually $uulength), chopping...");
      $uufile=substr($uufile,0,$MAX_FILE_LENGTH).".".$uuextension;
    }
    if (! -f "$scandir/$file_id/$file_id-$uuencoded_attachments-$uufile.uue") {
      open(UUIN,">$scandir/$file_id/$file_id-$uuencoded_attachments-$uufile.uue")||&tempfail("cannot open $scandir/$file_id/$file_id-$uuencoded_attachments-$uufile.uue - $!");
    } else {
      &tempfail("$scandir/$file_id/$file_id-$uuencoded_attachments-$uufile.uue already exists! - $!");
    }
    print UUIN "$begin 600 $uufile\n";
    print TMPFILE;
    while (<STDIN>) {
      print UUIN;
      print TMPFILE;
      if (/^end/) {
	close(UUIN)||&tempfail("cannot close $scandir/$file_id/$file_id-$uuencoded_attachments-$uufile - $!");
	#uudecode it - toss away the error code - who cares if it's broken...
	&debug("c_a_g_u: $uudecode_binary $uudecode_pipe $scandir/$file_id/$file_id-$uuencoded_attachments-$uufile.uue");
	open(UUOUT,"$uudecode_binary $uudecode_pipe $scandir/$file_id/$file_id-$uuencoded_attachments-$uufile.uue|");
	open(UUFILE,">$scandir/$file_id/$file_id-$uuencoded_attachments-$uufile");
	while (<UUOUT>) {
	  print UUFILE;
	}
	close UUOUT;
	close UUFILE;
	&debug("deleting uuencoded file as we have a decoded version of it now");
	unlink("$scandir/$file_id/$file_id-$uuencoded_attachments-$uufile.uue") if ($DEBUG < 100);
	last;
      }
    }
  }
}


sub log_msg() {};


###############################
#
##  END of standard subroutines
##  Virus-scanner specific subroutines automatically added below by setup.sh
##
###############################

#########################
## END of scanner definitions
##
#########################
